@@ -1,29 +1,33 @@
 (ns puppetlabs.services.ca.certificate-authority-core
-  (:import [java.io InputStream ByteArrayInputStream]
-           (clojure.lang IFn)
-           (org.joda.time DateTime))
-  (:require [puppetlabs.puppetserver.certificate-authority :as ca]
+  (:require [bidi.schema :as bidi-schema]
+            [cheshire.core :as cheshire]
+            [clj-time.core :as time]
+            [clj-time.format :as time-format]
+            [clojure.java.io :as io]
+            [clojure.string :as string]
+            [clojure.tools.logging :as log]
+            [liberator.core :refer [defresource]]
+            [liberator.representation :as representation]
+            [puppetlabs.comidi :as comidi :refer [ANY DELETE GET POST PUT]]
+            [puppetlabs.i18n.core :as i18n]
+            [puppetlabs.puppetserver.certificate-authority :as ca]
+            [puppetlabs.puppetserver.common :as common]
+            [puppetlabs.puppetserver.liberator-utils :as liberator-utils]
             [puppetlabs.puppetserver.ringutils :as ringutils]
             [puppetlabs.ring-middleware.core :as middleware]
             [puppetlabs.ring-middleware.utils :as middleware-utils]
-            [puppetlabs.puppetserver.liberator-utils :as liberator-utils]
-            [puppetlabs.comidi :as comidi :refer [GET ANY PUT DELETE]]
-            [bidi.schema :as bidi-schema]
-            [slingshot.slingshot :as sling]
-            [clojure.tools.logging :as log]
-            [clojure.java.io :as io]
-            [clojure.string :as string]
-            [clj-time.core :as time]
-            [clj-time.format :as time-format]
-            [schema.core :as schema]
-            [cheshire.core :as cheshire]
-            [liberator.core :refer [defresource]]
-            [liberator.representation :as representation]
+            [puppetlabs.ssl-utils.core :as utils]
+            [puppetlabs.trapperkeeper.authorization.ring :as ring]
+            [puppetlabs.trapperkeeper.authorization.ring-middleware :as auth-middleware]
+            [puppetlabs.trapperkeeper.services.status.status-core :as status-core]
             [ring.util.request :as request]
             [ring.util.response :as rr]
-            [puppetlabs.trapperkeeper.services.status.status-core :as status-core]
-            [puppetlabs.i18n.core :as i18n]
-            [puppetlabs.ssl-utils.core :as utils]))
+            [schema.core :as schema]
+            [slingshot.slingshot :refer [try+]])
+  (:import (clojure.lang IFn)
+           (java.io ByteArrayInputStream InputStream StringWriter)
+           (java.security.cert X509Certificate)
+           (org.joda.time DateTime)))
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;; Constants
@@ -34,9 +38,6 @@
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;; 'handler' functions for HTTP endpoints
 
-;; Perhaps this could/should be in the tk context somewhere, and of
-;; course it'll need to be used to guard any competing writes.
-(def crl-write-serializer (Object.))
 
 (schema/defn format-http-date :- (schema/maybe DateTime)
   "Formats an http-date into joda time.  Returns nil for malformed or nil
@@ -71,61 +72,70 @@
         (rr/not-found (i18n/tru "Could not find certificate_request {0}" subject)))
       (rr/content-type "text/plain")))
 
+
 (schema/defn handle-put-certificate-request!
-  [subject :- String
-   certificate-request :- InputStream
-   ca-settings :- ca/CaSettings]
-  (sling/try+
-    (ca/process-csr-submission! subject certificate-request ca-settings)
-    (rr/content-type (rr/response nil) "text/plain")
+  [ca-settings :- ca/CaSettings
+   report-activity
+   {:keys [body] {:keys [subject]} :route-params :as request}]
+  (try+
+    (let [report-activity-fn (ca/create-report-activity-fn report-activity request)]
+      (ca/process-csr-submission! subject body ca-settings report-activity-fn)
+      (rr/content-type (rr/response nil) "text/plain"))
     (catch ca/csr-validation-failure? {:keys [msg]}
       (log/error msg)
       ;; Respond to all CSR validation failures with a 400
       (middleware-utils/plain-response 400 msg))))
 
 
+(schema/defn resolve-crl-information
+  "Create a map that has the appropriate path, lock, timeout and descriptor for the crl being used"
+  [{:keys [enable-infra-crl cacrl infra-crl-path crl-lock crl-lock-timeout-seconds]} :- ca/CaSettings]
+  {:path (if (true? enable-infra-crl) infra-crl-path cacrl)
+   :lock crl-lock
+   :descriptor ca/crl-lock-descriptor
+   :timeout crl-lock-timeout-seconds})
+
 (defn handle-get-certificate-revocation-list
   "Always return the crl if no 'If-Modified-Since' header is provided or
   if that header is not in correct http-date format. If the header is
-  present and has correct format, only return the crl if the master
+  present and has correct format, only return the crl if the server
   cacrl is newer than the agent crl."
-  [request {:keys [cacrl infra-crl-path enable-infra-crl]}]
+  [request ca-settings]
   (let [agent-crl-last-modified-val (rr/get-header request "If-Modified-Since")
         agent-crl-last-modified-date-time (format-http-date agent-crl-last-modified-val)
-        master-crl-to-use (if (true? enable-infra-crl) infra-crl-path cacrl)
-        master-crl-last-modified-date-time (ca/get-file-last-modified master-crl-to-use)]
-    (if (or (nil? agent-crl-last-modified-date-time)
-            (time/after? master-crl-last-modified-date-time agent-crl-last-modified-date-time))
-      (-> (ca/get-certificate-revocation-list master-crl-to-use)
-          (rr/response)
-          (rr/content-type "text/plain"))
-      (-> (rr/response nil)
-          (rr/status 304)
-          (rr/content-type "text/plain")))))
+        {:keys [path lock descriptor timeout]} (resolve-crl-information ca-settings)]
+    ;; Since the locks are reentrant, obtain the read lock to prevent modification during the
+    ;; window of time between when the last-modified is read and when the crl content is potentially read.
+    (common/with-safe-read-lock lock descriptor timeout
+        (if (or (nil? agent-crl-last-modified-date-time)
+                (time/after? (ca/get-file-last-modified path lock descriptor timeout)
+                             agent-crl-last-modified-date-time))
+          (-> (ca/get-certificate-revocation-list path lock descriptor timeout)
+              (rr/response)
+              (rr/content-type "text/plain"))
+          (-> (rr/response nil)
+              (rr/status 304)
+              (rr/content-type "text/plain"))))))
 
 (schema/defn handle-put-certificate-revocation-list!
   [incoming-crl-pem :- InputStream
-   {:keys [cacrl cacert enable-infra-crl infra-crl-path]} :- ca/CaSettings]
-  (locking crl-write-serializer
-    (try
-      (let [byte-stream (-> incoming-crl-pem
-                            ca/input-stream->byte-array
-                            ByteArrayInputStream.)
-            incoming-crls (utils/pem->crls byte-stream)]
-        (if (empty? incoming-crls)
-          (do
-            (log/info (i18n/trs "No valid CRLs submitted, nothing will be updated."))
-            (middleware-utils/plain-response 400 "No valid CRLs submitted."))
-
-          (do
-            (ca/update-crls incoming-crls cacrl cacert)
-            (when enable-infra-crl
-              (ca/update-crls incoming-crls infra-crl-path cacert))
-            (middleware-utils/plain-response 200 "Successfully updated CRLs."))))
-      (catch IllegalArgumentException e
-        (let [error-msg (.getMessage e)]
-          (log/error error-msg)
-          (middleware-utils/plain-response 400 error-msg))))))
+   {:keys [cacrl cacert] :as ca-settings} :- ca/CaSettings]
+  (try
+    (let [byte-stream (-> incoming-crl-pem
+                          ca/input-stream->byte-array
+                          ByteArrayInputStream.)
+          incoming-crls (utils/pem->crls byte-stream)]
+      (if (empty? incoming-crls)
+        (do
+          (log/info (i18n/trs "No valid CRLs submitted, nothing will be updated."))
+          (middleware-utils/plain-response 400 "No valid CRLs submitted."))
+        (do
+          (ca/update-crls! incoming-crls cacrl cacert ca-settings)
+          (middleware-utils/plain-response 200 "Successfully updated CRLs."))))
+    (catch IllegalArgumentException e
+      (let [error-msg (.getMessage e)]
+        (log/error error-msg)
+        (middleware-utils/plain-response 400 error-msg)))))
 
 (schema/defn handle-delete-certificate-request!
   [subject :- String
@@ -154,10 +164,17 @@
     (catch Exception e
       (log/debug e))))
 
+(schema/defn certificate-issued? :- schema/Bool
+  [settings :- ca/CaSettings
+   subject :- schema/Str]
+  (or (ca/certificate-exists? settings subject)
+      (ca/in-cert-inventory-file? settings subject)))
+
 (schema/defn handle-cert-clean
-  [{:keys [body]}
-   ca-settings :- ca/CaSettings]
-  (if-let [json-body (try-to-parse body)]
+  [request
+   ca-settings :- ca/CaSettings
+   report-activity]
+  (if-let [json-body (try-to-parse (:body request))]
     ;; TODO support async mode
     (if (true? (:async json-body))
       (-> (rr/response "Async mode is not currently supported.")
@@ -166,18 +183,20 @@
       (if-let [certnames (:certnames json-body)]
         (let [{existing-certs true
                missing-certs false} (group-by
-                                     #(ca/certificate-exists? ca-settings %)
+                                     #(certificate-issued? ca-settings %)
                                      certnames)
               message (when (seq missing-certs)
                         (format "The following certs do not exist and cannot be revoked: %s"
-                                (vec missing-certs)))]
+                                (vec missing-certs)))
+              report-activity-fn (ca/create-report-activity-fn report-activity request)]
           (try
-            (ca/revoke-existing-certs! ca-settings existing-certs)
+            (ca/revoke-existing-certs! ca-settings existing-certs report-activity-fn)
             (ca/delete-certificates! ca-settings existing-certs)
             (-> (rr/response (or message "Successfully cleaned all certs."))
                 (rr/status 200)
                 (rr/content-type "text/plain"))
             (catch Exception e
+              (log/info e (i18n/trs "Error while cleaning certs"))
               (-> (rr/response (str "Error while cleaning certs: " (.getMessage e)))
                   (rr/status 500)
                   (rr/content-type "text/plain")))))
@@ -188,6 +207,119 @@
         (rr/status 400)
         (rr/content-type "text/plain"))))
 
+(schema/defn validate-cert-in-infra-list :- schema/Bool
+  [request-cert :- X509Certificate
+   infra-nodes-path :- schema/Str]
+  (if infra-nodes-path
+    (let [subject-name (utils/get-cn-from-x509-certificate request-cert)]
+      (if (.exists (io/file infra-nodes-path))
+        (with-open [infra-nodes-reader (io/reader infra-nodes-path)]
+          (let [infra-nodes (ca/read-infra-nodes infra-nodes-reader)]
+            (boolean (some #(= subject-name %) infra-nodes))))
+        (do (log/info (i18n/trs "Unable to find infra-nodes file at {0}" infra-nodes-path))
+            false)))
+    (do
+      (log/info (i18n/trs "infra-nodes-path is nil, cannot validate certificate is allowed."))
+      false)))
+
+(schema/defn validate-header-cert-not-revoked :- (schema/maybe X509Certificate)
+  "Given a certificate, validate that the certificate is not in the CRL. The
+  messaging is specific to the header method of certificate delivery. If the
+  certificate is valid, it is returned, otherwise return nil"
+  [cert :- X509Certificate
+   ca-settings :- ca/CaSettings]
+  (if-not (ca/is-revoked? cert ca-settings)
+    cert
+    (log/warn (i18n/trs "Rejecting certificate request because the {0} was specified and the certificate is revoked." auth-middleware/header-cert-name))))
+
+(schema/defn request->cert :- (schema/maybe X509Certificate)
+  "Pull the client certificate from the request.  Response includes the
+  certificate as a java.security.cert.X509Certificate object or, if none
+  can be found, nil.  allow-header-cert-info determines whether to try to
+  pull the certificate from an HTTP header (true) or from the certificate
+  provided during SSL session negotiation (false).
+
+  If allow-header-cert-info is false, and the cert is present in both the header
+  and the request, validate that the cert in the request is in the infra list.
+  If it isn't in the infra list, log the issue and return nil.
+  If the header isn't set, return the cert from the request.
+  "
+  [request :- ring/Request
+   {:keys [allow-header-cert-info infra-nodes-path] :as ca-settings} :- ca/CaSettings]
+  (let [header-cert-val (get-in request [:headers auth-middleware/header-cert-name])]
+    (if allow-header-cert-info
+      (validate-header-cert-not-revoked (auth-middleware/header->cert header-cert-val) ca-settings)
+      (if-let [request-cert (:ssl-client-cert request)]
+        (if header-cert-val
+          (if (validate-cert-in-infra-list request-cert infra-nodes-path)
+            (validate-header-cert-not-revoked (auth-middleware/header->cert header-cert-val) ca-settings)
+            (log/warn (i18n/trs "Rejecting certificate request because the {0} header was specified, but the client making the request was not in the allow list."  auth-middleware/header-cert-name)))
+          request-cert)
+        (log/warn (i18n/trs "Request is missing a certificate for an endpoint that requires a certificate."))))))
+
+(schema/def Certnames [schema/Str])
+
+(schema/defn handle-bulk-cert-signing
+  [request
+   ca-settings :- ca/CaSettings
+   report-activity]
+  (let [json-body (try-to-parse (:body request)) 
+        certnames (:certnames json-body)] 
+    (if-let [schema-error (schema/check Certnames certnames)]
+      (-> (rr/response (cheshire/generate-string {:kind :schema-violation
+                                                  :submitted certnames
+                                                  :error (str schema-error)}))
+          (rr/status 422)
+          (rr/content-type "application/json"))
+      (-> (rr/response (cheshire/generate-string (ca/sign-multiple-certificate-signing-requests! certnames ca-settings report-activity)))
+          (rr/status 200)
+          (rr/content-type "application/json")))))
+
+(schema/defn handle-bulk-cert-signing-all
+  [ca-settings :- ca/CaSettings report-activity]
+  (let [csr-files (-> (ca/get-paths-to-all-certificate-requests (:csrdir ca-settings))
+                      (common/extract-file-names-from-paths)
+                      (common/remove-suffix-from-file-names ".pem"))
+        results (ca/sign-multiple-certificate-signing-requests! csr-files ca-settings report-activity)]
+    (-> (rr/response (cheshire/generate-string results))
+        (rr/status 200)
+        (rr/content-type "application/json"))))
+
+(schema/defn ^:always-validate
+  handle-cert-renewal
+  "Given a request and the CA settings, if there is a cert present in the request
+  (either in the ssl-client-cert property of the request, or as an x-client-cert
+  field in the header when allow-header-cert-info is set to true) and the cert in
+  the request is valid and signed by the this CA. then generate a renewed cert and
+  return it in the response body"
+  [request
+   {:keys [cacert cakey allow-auto-renewal] :as ca-settings} :- ca/CaSettings
+   report-activity]
+  (if allow-auto-renewal
+    (if-let [request-cert (request->cert request ca-settings)]
+      (let [signing-cert (utils/pem->ca-cert cacert cakey)]
+        (if (ca/cert-authority-id-match-ca-subject-id? request-cert signing-cert)
+          (do
+            (log/info (i18n/trs "Certificate present, processing renewal request"))
+            (let [cert-signing-result (ca/renew-certificate! request-cert ca-settings report-activity)
+                  cert-writer (StringWriter.)]
+              ;; has side effect of writing to the writer
+              (utils/cert->pem! cert-signing-result cert-writer)
+              (-> (rr/response (.toString cert-writer))
+                  (rr/content-type "text/plain"))))
+          (do
+            (log/info (i18n/trs "Certificate present, but does not match signature"))
+            (-> (rr/response (i18n/tru "Certificate present, but does not match signature"))
+                (rr/status 403)
+                (rr/content-type "text/plain")))))
+        (do
+          (log/info (i18n/trs "No valid certificate found in renewal request"))
+          (-> (rr/bad-request (i18n/tru "No valid certificate found in renewal request"))
+              (rr/content-type "text/plain"))))
+    (-> (rr/response "Not Found")
+        (rr/status 404)
+        (rr/content-type "text/plain"))))
+
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;; Web app
 
@@ -198,7 +330,7 @@
   [true {::malformed message}])
 
 (defn conflict
-  "Returns a value indicating to liberator that the request is is conflict
+  "Returns a value indicating to liberator that the request is conflict
   with the server, with the given error message assoc'ed into the context."
   [message]
   [true {::conflict message}])
@@ -274,7 +406,7 @@
     (representation/ring-response)))
 
 (defresource certificate-status
-  [subject settings]
+  [subject settings report-activity]
   :allowed-methods [:get :put :delete]
 
   :available-media-types media-types
@@ -287,7 +419,7 @@
      (case desired-state
        :revoked
        ;; A signed cert must exist if we are to revoke it.
-       (when-not (ca/certificate-exists? settings subject)
+       (when-not (certificate-issued? settings subject)
          (conflict (i18n/tru "Cannot revoke certificate for host {0} without a signed certificate" subject)))
 
        :signed
@@ -308,8 +440,8 @@
   :exists?
   (fn [_context]
     (or
-     (ca/certificate-exists? settings subject)
-     (ca/csr-exists? settings subject)))
+      (certificate-issued? settings subject)
+      (ca/csr-exists? settings subject)))
 
   :handle-conflict
   (fn [context]
@@ -377,12 +509,14 @@
 
   :put!
   (fn [context]
-     (let [desired-state (get-desired-state context)]
-       (locking crl-write-serializer
-         (ca/set-certificate-status!
-          (merge-request-settings settings context)
-          subject
-          desired-state))
+     (let [desired-state (get-desired-state context)
+           request (:request context)
+           report-activity-fn (ca/create-report-activity-fn report-activity request)]
+       (ca/set-certificate-status!
+         (merge-request-settings settings context)
+         subject
+         desired-state
+         report-activity-fn)
        (-> context
          (assoc-in [:representation :media-type] "text/plain")))))
 
@@ -406,11 +540,12 @@
         (as-json-or-pson context)))))
 
 (schema/defn ^:always-validate web-routes :- bidi-schema/RoutePair
-  [ca-settings :- ca/CaSettings]
+  [ca-settings :- ca/CaSettings
+   report-activity]
   (comidi/routes
     (comidi/context ["/v1"]
       (ANY ["/certificate_status/" :subject] [subject]
-        (certificate-status subject ca-settings))
+          (certificate-status subject ca-settings report-activity))
       (comidi/context ["/certificate_statuses/"]
         (ANY [[#"[^/]+" :ignored-but-required]] request
           (certificate-statuses request ca-settings))
@@ -420,8 +555,8 @@
       (comidi/context ["/certificate_request/" :subject]
         (GET [""] [subject]
           (handle-get-certificate-request subject ca-settings))
-        (PUT [""] [subject :as {body :body}]
-          (handle-put-certificate-request! subject body ca-settings))
+        (PUT [""] request
+          (handle-put-certificate-request! ca-settings report-activity request))
         (DELETE [""] [subject]
           (handle-delete-certificate-request! subject ca-settings)))
       (GET ["/certificate_revocation_list/" :ignored-node-name] request
@@ -431,7 +566,13 @@
       (GET ["/expirations"] _request
         (handle-get-ca-expirations ca-settings))
       (PUT ["/clean"] request
-        (handle-cert-clean request ca-settings)))
+        (handle-cert-clean request ca-settings report-activity))
+      (POST ["/certificate_renewal"] request
+        (handle-cert-renewal request ca-settings report-activity))
+      (POST ["/sign"] request
+        (handle-bulk-cert-signing request ca-settings report-activity))
+      (POST ["/sign/all"] _request
+        (handle-bulk-cert-signing-all  ca-settings report-activity)))
     (comidi/not-found "Not Found")))
 
 (schema/defn ^:always-validate
